// This array makes sure every player has a
// FollowerControl item in their inventory.
class FollowerSpawningHandler : EventHandler
{
	override void PlayerEntered(playerEvent e)
	{
		// First, don't do anything if it's a TITLEMAP,
		// or the player is invalid, or doesn't have
		// a PlayerPawn attached for some reason:
		if (level.Mapname == "TITLEMAP")
			return;
		if (!PlayerInGame[e.PlayerNumber])
			return;
		let plr = players[e.PlayerNumber].mo;
		if (!plr)
			return;
		// Check if the player has the control item
		// in their inventory:
		let cont = FollowerControl(plr.FindInventory("FollowerControl"));
		// If they do, respawn their followers:
		if (cont)
			cont.RespawnFollowers();
		// If they don't, give them the item:
		else
			plr.GiveInventory("FollowerControl", 1);		
	}
}

class BaseFollower : /*SwitchableDecoration*/Actor
{
	bool created;
	
	Default
	{
		+SOLID
		+USESPECIAL
		Activation THINGSPEC_Activate;
	}
	
	override void PostBeginPlay()
	{
		super.PostBeginPlay();
		if (!master)
		{
			FollowerControl.BindFollower(self);
		}
	}
}

// This is a data container that contains the data
// for the follower. The 'mo' and 'type' fields are
// mandatory. More fields can be added, if necessary;
// in that case they will need to be added to the 
// SpawnFollower() function defined in FollowerControl
// (see further below).
class FollowerData play
{
	Actor mo;
	class<Actor> type;
	int health;
}

// This item must be placed in the player's inventory,
// which is done via FollowerSpawningHandler.
// This item stores an array of FollowerData instances,
// thus saving all follower-related data and transferring
// it between maps (since items don't get stripped).
class FollowerControl : Inventory
{
	array <FollowerData> followers;
	
	Default
	{
		+INVENTORY.UNDROPPABLE
		+INVENTORY.UNTOSSABLE
		Inventory.maxamount 1;
	}
	
	// This is the function necessary to initially create
	// a follower that didn't exist before.
	// The first argument must be a pointer to the PlayerPawn,
	// the second--the name of the follower class:
	static void CreateFollower(Actor ppawn, class<Actor> follower, vector3 pos = (0,0,0))
	{
		if (!ppawn || !ppawn.player)
			return;
		// Find the control item in player's inventory:
		let cont = FollowerControl(ppawn.FindInventory("FollowerControl"));
		if (!cont)
			return;
		// Create a new instance of FollowerData:
		let fdata = New("FollowerData");
		if (!fdata)
		{
			console.printf("Error: couldn't create an instance of FollowerData");
			return;
		}
		// Spawn a follower:
		cont.SpawnFollower(fdata, follower, ppawn, pos);
		// Record the related instance of FollowerData
		// in the array in FollowerControl:
		cont.followers.Push(fdata);
	}
	
	static bool BindFollower(Actor follower, int pnumber = 0)
	{
		if (!follower)
		{
			console.printf("Error: tried to bind a nonexistent follower");
			return false;
		}
		if (!PlayerInGame[pnumber])
		{		
			console.printf("Error: player %d is not in the game", pnumber);
			return false;
		}
		let ppawn = players[pnumber].mo;
		if (!ppawn)
		{		
			console.printf("Error: player %d does not have a PlayerPawn attached", pnumber);
			return false;
		}
		let cont = FollowerControl(ppawn.FindInventory("FollowerControl"));
		if (!cont)
		{		
			console.printf("Error: player %d does not have a FollowerControl instance in their inventory", pnumber);
			return false;
		}
		// Create a new instance of FollowerData:
		let fdata = New("FollowerData");
		if (!fdata)
		{
			console.printf("Error: couldn't create an instance of FollowerData");
			return false;
		}
		follower.bFRIENDLY = true; //make it friendly		
		follower.master = ppawn;
		fdata.mo = follower; //save a pointer to it in FollowerData
		fdata.type = follower.GetClass(); //record its class type
		// Record the related instance of FollowerData
		// in the array in FollowerControl:
		cont.followers.Push(fdata);
		return true;
	}
	
	// This is a base follower-spawning function.
	// It's not meant to be used directly; instead
	// it's used by CreateFollower() and RespawnFollowers()
	// to spawn and record a follower when needed.
	protected void SpawnFollower(FollowerData fdata, class<Actor> follower, Actor ppawn, vector3 pos = (0,0,0))
	{
		if (!ppawn)	
			return;
		if (!fdata)
		{
			console.printf("Error: invalid FollowerData reference passed. Spawning follower canceleld");
			return;
		}
		vector3 targetpos = (pos == (0,0,0)) ? ppawn.pos : pos;
		// Spawn the follower:
		let flwr = Spawn(follower, targetpos);
		if (!flwr)
		{
			console.printf("Error: couldn't spawn %s at %.1f.%.1f.%.1f", follower.GetClassName(), ppawn.pos.x, ppawn.pos.y, ppawn.pos.z);
			return;
		}
		flwr.bFRIENDLY = true; //make it friendly	
		flwr.master = ppawn;
		fdata.mo = flwr; //save a pointer to it in FollowerData
		fdata.type = follower; //record its class type
		
		// This is a debug function. Uncomment to get info at follower spawn:
		
		//console.printf("Spawned %s at at %.1f.%.1f.%.1f with %d health", follower.GetClassName(),ppawn.pos.x, ppawn.pos.y, ppawn.pos.z, fdata.health);
	}		
	
	// This function reads the FollowerData array
	// and respawns followers based on the 'type'
	// field. It's meant to be called by the 
	// event handler:
	void RespawnFollowers()
	{
		if (!owner)
			return;
		// Iterater through the array:
		for (int i = 0; i < followers.Size(); i++)
		{
			if (!followers[i])
				continue;
			// Get the 'type' (class name) of the 
			// follower we need to spawn:
			let type = followers[i].type;
			if (!type)
			{
				console.printf("Error: index %d of the followers array has no type", i);
				followers.Delete(i);
				continue;
			}
			// Spawn the follower and record it again:
			SpawnFollower(followers[i], type, owner);
		}
	}
	
	// This is to improve performance: this control
	// item can't exist in the world, so it doesn't
	// need a Tick():
	override void Tick() {}
	
	// This is called every tic while the item is
	// in the player's inventory to make sure
	// FollowerData tracks every follower's health:
	override void DoEffect()
	{
		super.DoEffect();
		
		if (!owner || !owner.player)
			return;
		
		for (int i = 0; i < followers.Size(); i++)
		{
			if (!followers[i])
				continue;
			let mo = followers[i].mo;
			if (!mo)
				continue;
			// If the follower is dead, remove its data
			// from the array:
			if (mo.health <= 0)
			{
				followers.Delete(i);
				continue;
			}
			followers[i].health = mo.health;
		}
	}
}


// This is an example basic follower spawner. 
// It spawns and record a follower once you press 
// Use in front of it:
class FollowerSpawner : Actor
{
	Default
	{
		+SOLID
		+USESPECIAL
		Activation THINGSPEC_Activate;
		Renderstyle "Translucent";
	}
	
	override void Activate(actor activator)
	{
		// Check that activator is valid and is a player:
		if (!activator || !activator.player)
			return;
		// Spawn an Imp follower at the player's position:
		FollowerControl.CreateFollower(activator, "DoomImp", pos);
		SetStateLabel("Death"); //the spawner disappears afterwards
	}
	
	States
	{
	Spawn:
		MEGA A -1 bright;
		stop;
	Death:
		#### # 1 A_FadeOut(0.1);
		loop;
	}
}
